# Copyright (C) 2019 Arturs Ontuzans
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import bpy, math

class Rig_OT_Operator(bpy.types.Operator):
    bl_idname = "view3d.rig_vehicle"
    bl_label = "Rig Vehicle"
    bl_description = "Rigs Vehicle for UE4"
    bl_options = {'REGISTER', 'UNDO'}

    def set_vertex_group(self, mesh, vertex_group):
        #Make a new list for all vertice indexes
        index_list = [0]*len(mesh.data.vertices)
        #Populate index list with vertice indexes
        mesh.data.vertices.foreach_get('index', index_list)   
        #Set mesh vertex group
        mesh.vertex_groups[vertex_group].add(index_list, 1, 'REPLACE')
    
    def add_child_bone(self, bone_name, parent_bone, wheel_mesh, armature_data):
        #Create a new bone
        new_bone = armature_data.data.edit_bones.new(bone_name)
        #Set bone's size
        new_bone.head = (0,0,0)
        new_bone.tail = (0,100,0)
        #Set bone's parent
        new_bone.parent = parent_bone
        #Set bone's location to wheel
        new_bone.matrix = wheel_mesh.matrix_world
        return new_bone

    @classmethod
    def poll(cls, context):
        unit_length = bpy.context.scene.unit_settings.scale_length
        unit_system = bpy.context.scene.unit_settings.system
        return (context.scene.vehicle_base is not None 
            and context.scene.wheel_FR is not None 
            and context.scene.wheel_FL is not None 
            and context.scene.wheel_RR is not None 
            and context.scene.wheel_RL is not None 
            and math.isclose(unit_length, 0.01, abs_tol=0.001)
            and unit_system == 'METRIC')

    def execute(self, context):
        scene = context.scene

        C = bpy.context 
        D = bpy.data
        O = bpy.ops

        #Set variables for vehicle base and wheel meshes
        vehicle_base = scene.vehicle_base
        wheel_RL = scene.wheel_RL
        wheel_RR = scene.wheel_RR
        wheel_FL = scene.wheel_FL
        wheel_FR = scene.wheel_FR

        #Set vehicle base as active
        C.view_layer.objects.active = vehicle_base

        #Select all vehicle meshes
        vehicle_base.select_set(state=True)
        wheel_RL.select_set(state=True)
        wheel_RR.select_set(state=True)
        wheel_FL.select_set(state=True)
        wheel_FR.select_set(state=True)

        #Set object mode
        O.object.mode_set(mode='OBJECT', toggle=True)

        #Set all object origins to geometry center
        O.object.origin_set(type='ORIGIN_GEOMETRY', center = 'MEDIAN')

        #Apply object transform
        O.object.transform_apply(location = False, rotation = True, scale = True)

        #Create armature object
        armature = D.armatures.new('Armature')
        armature_object = D.objects.new('Armature', armature)

        #Link armature object to our scene
        C.collection.objects.link(armature_object)

        #Make armature variable
        armature_data = D.objects[armature_object.name]

        #Set armature active
        C.view_layer.objects.active = armature_data

        #Set armature selceted
        armature_data.select_set(state=True)

        #Set edit mode
        O.object.mode_set(mode='EDIT', toggle=False)

        #Set bones In front and show axis
        armature_data.show_in_front = True
        armature_data.data.show_axes = True

        #Add root bone
        root_bone = armature_data.data.edit_bones.new('Root')
        #Set its orientation and size
        root_bone.head = (0,0,0)
        root_bone.tail = (0,100,0)
        #Set its location to vehicle base mesh
        root_bone.matrix = vehicle_base.matrix_world

        #Add wheel bones to armature
        self.add_child_bone('RL', root_bone, wheel_RL, armature_data)
        self.add_child_bone('RR', root_bone, wheel_RR, armature_data)
        self.add_child_bone('FL', root_bone, wheel_FL, armature_data)
        self.add_child_bone('FR', root_bone, wheel_FR, armature_data)

        #Set object mode
        O.object.mode_set(mode='OBJECT', toggle=True)

        #Select vehicle meshes
        vehicle_base.select_set(state=True)
        wheel_RL.select_set(state=True)
        wheel_RR.select_set(state=True)
        wheel_FL.select_set(state=True)
        wheel_FR.select_set(state=True)

        #Set armature active
        C.view_layer.objects.active = armature_data

        #Parent meshes to armature with empty groups
        O.object.parent_set(type='ARMATURE_NAME')

        #Set mesh vertex groups/weightpaint mesh
        self.set_vertex_group(vehicle_base, 'Root')
        self.set_vertex_group(wheel_RL, 'RL')
        self.set_vertex_group(wheel_RR, 'RR')
        self.set_vertex_group(wheel_FL, 'FL')
        self.set_vertex_group(wheel_FR, 'FR')

        #Deselect all objects
        O.object.select_all(action='DESELECT')
        #Set pose mode
        O.object.mode_set(mode='POSE', toggle=True)

        return {'FINISHED'}


class Scale_Units_OT_Operator(bpy.types.Operator):
    bl_idname = "view3d.set_unit_scale"
    bl_label = "Set Unit Scale"
    bl_description = "Set Scene Unit Scale for UE4"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        unit_length = bpy.context.scene.unit_settings.scale_length
        return not math.isclose(unit_length, 0.01, abs_tol=0.001) or bpy.context.scene.unit_settings.system != 'METRIC'

    def execute(self, context):
        bpy.context.scene.unit_settings.system = 'METRIC'
        bpy.context.scene.unit_settings.scale_length = 0.01
        bpy.context.space_data.clip_end = 100000

        return {'FINISHED'}


class Upscale_Objects_OT_Operator(bpy.types.Operator):
    bl_idname = "view3d.upscale_objects"
    bl_label = "Rig Vehicle"
    bl_description = "Set Scene Unit Scale for UE4"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        bpy.ops.object.mode_set(mode='OBJECT', toggle=True)

        bpy.ops.view3d.snap_cursor_to_center()
        bpy.ops.object.select_all(action='SELECT')

        bpy.context.scene.tool_settings.transform_pivot_point = 'CURSOR'

        bpy.ops.transform.resize(value=(100, 100, 100), orient_type='GLOBAL', 
            orient_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)), 
            orient_matrix_type='GLOBAL', mirror=True, use_proportional_edit=False, 
            proportional_edit_falloff='SMOOTH', proportional_size=1, 
            use_proportional_connected=False, use_proportional_projected=False)

        bpy.ops.object.transform_apply(location = False, rotation = True, scale = True)

        bpy.context.scene.tool_settings.transform_pivot_point = 'MEDIAN_POINT'
        
        return {'FINISHED'}